# coding=utf-8
from django_multisafepay.data import CustomerDelivery, Merchant, Transaction
from django_multisafepay.data.status import CheckoutData, CustomerStatus, Ewallet, PaymentDetails, TransactionStatus

from .base import XmlRequest, XmlResponse


class Status(XmlRequest):
    """
    The message to request a status
    """
    xml_name = 'status'
    xml_fields = (
        'merchant',
        'transaction',
    )

    def __init__(self, merchant, transaction_id):
        """
        :param merchant: The merchant, with the fields account, site_id, site_code filled in.
        :type merchant: Merchant
        """
        self.merchant = Merchant(
            # Only these fields are required:
            account=merchant.account,
            site_id=merchant.site_id,
            site_code=merchant.site_secure_code,
        )
        self.transaction = Transaction(id=transaction_id)


class StatusReply(XmlResponse):
    """
    Reply from a status call.
    """

    # Fast Checkout example:
    # -----------------------
    #
    # <status result="ok">
    #   <ewallet>
    #     <id>2118132</id>
    #     <status>completed</status>
    #     <fastcheckout>YES</fastcheckout>
    #     <created>20140305191332</created>
    #     <modified>20140305191335</modified>
    #     <reasoncode/>
    #     <reason/>
    #   </ewallet>
    #   <customer>
    #     <amount>2000</amount>      <!-- NOTE these 3 extra fields!! -->
    #     <currency>EUR</currency>
    #     <account/>
    #     <locale>en_US</locale>
    #     <firstname>Diederik</firstname>
    #     <lastname>van der Boor</lastname>
    #     <address1>Foo</address1>
    #     <address2/>
    #     <housenumber>...</housenumber>
    #     <zipcode>...</zipcode>
    #     <city>...</city>
    #     <state/>
    #     <country>NL</country>
    #     <countryname/>
    #     <phone1/>
    #     <phone2/>
    #     <email>foo@example.org</email>
    #   </customer>
    #   <customer-delivery/>
    #   <transaction>
    #     <id>7</id>
    #     <currency>EUR</currency>
    #     <amount>2000</amount>
    #     <description>...</description>
    #     <var1>5</var1>
    #     <var2/>
    #     <var3/>
    #     <items/>
    #   </transaction>
    #   <paymentdetails>
    #     <type>IDEAL</type>
    #     <accountiban>NL30INGB0123456789</accountiban>     <!--  NOTE these 3 extra fields because of iDEAL!! -->
    #     <accountbic>INGBNL2A</accountbic>
    #     <accountid>654412345</accountid>
    #     <accountholdername>Mr Küppers</accountholdername>
    #     <externaltransactionid>0050000081927015</externaltransactionid>
    #   </paymentdetails>
    #   <checkoutdata version="0.1">
    #     <checkout-flow-support><merchant-checkout-flow-support>
    #       <shipping-methods>
    #         <pickup name="Online">
    #           <price currency="EUR">0.00</price>
    #         </pickup>
    #       </shipping-methods>
    #     </merchant-checkout-flow-support></checkout-flow-support>
    #     <order-adjustment>
    #       <shipping>
    #         <pickup>
    #           <shipping-name>Online</shipping-name>           <!--  NOTE different tag format!! -->
    #           <shipping-cost currency="EUR">0.00</shipping-cost>
    #         </pickup>
    #       </shipping>
    #       <adjustment-total currency="EUR">0.00</adjustment-total>
    #       <total-tax currency="EUR">0.00</total-tax>
    #     </order-adjustment>
    #     <order-total currency="EUR">20.00</order-total>
    #   </checkoutdata>
    # </status>
    #
    #
    # Connect example:
    # ------------------
    #
    # <status result="ok">
    #   <ewallet>
    #     <id>50102723</id>
    #     <status>completed</status>
    #     <fastcheckout>NO</fastcheckout>
    #     <created>20150526130908</created>
    #     <modified>20150526131108</modified>
    #     <reasoncode />
    #     <reason />
    #   </ewallet>
    #   <customer>
    #     <amount>6000</amount>
    #     <currency>EUR</currency>
    #     <account />
    #     <locale>nl_NL</locale>
    #     <firstname>Diederik</firstname>
    #     <lastname>van der Boor</lastname>
    #     <address1>Foo</address1>
    #     <address2 />
    #     <housenumber>...</housenumber>
    #     <zipcode>...</zipcode>
    #     <city>...</city>
    #     <state />
    #     <country>FR</country>
    #     <countryname />
    #     <phone1>+33 123456789</phone1>
    #     <phone2 />
    #     <email>foo@example.org</email>
    #   </customer>
    #   <customer-delivery />
    #   <transaction>
    #     <id>10217</id>
    #     <recurringid />
    #     <currency>EUR</currency>
    #     <amount>6000</amount>
    #     <cost>174</cost>
    #     <description>...</description>
    #     <var1>215</var1>
    #     <var2 />
    #     <var3 />
    #     <items />
    #     <amountrefunded>0</amountrefunded>
    #   </transaction>
    #   <paymentdetails>
    #     <type>MASTERCARD</type>
    #     <accountid />
    #     <accountholdername>A.B. Tester/accountholdername>
    #     <externaltransactionid>2-70-370907</externaltransactionid>
    #   </paymentdetails>
    # </status>

    def __init__(self, ewallet, customer, customer_delivery, transaction, payment_details, checkoutdata=None):
        """
        :type ewallet: Ewallet
        :type customer: CustomerStatus
        :type customer_delivery: CustomerDelivery
        :type transaction: Transaction
        :type payment_details: PaymentDetails
        :type checkoutdata: CheckoutData
        """
        self.ewallet = ewallet
        self.customer = customer
        self.customer_delivery = customer_delivery
        self.transaction = transaction
        self.payment_details = payment_details
        self.checkoutdata = checkoutdata

    STATUS_INITIALIZED = "initialized"  # waiting
    STATUS_COMPLETED = "completed"     # payment complete
    STATUS_WAITING = "uncleared"       # waiting (credit cards or direct debit)
    STATUS_CANCELLED = "void"          # canceled
    STATUS_DECLINED = "declined"       # declined
    STATUS_REFUNDED = "refunded"       # refunded
    STATUS_EXPIRED = "expired"         # expired

    @classmethod
    def get_class_kwargs(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        return dict(
            ewallet=Ewallet.from_xml(xml.find('ewallet')),
            customer=CustomerStatus.from_xml(xml.find('customer')),
            customer_delivery=CustomerDelivery.from_xml(xml.find('customer-delivery')),
            transaction=TransactionStatus.from_xml(xml.find('transaction')),
            payment_details=PaymentDetails.from_xml(xml.find('paymentdetails')),
            checkoutdata=CheckoutData.from_xml(xml.find('checkoutdata')),
        )

    @property
    def status_code(self):
        return self.ewallet.status



"""
The data objects can be provided as parameters to the client.
"""
from .base import Price
from .merchant import Merchant, Plugin
from .customer import Customer, CustomerDelivery
from .gateway import GatewayCustomer, Gateway
from .gatewayinfo import GatewayInfo
from .transaction import Transaction, CheckoutSettings, GoogleAnalytics
from .cart import Cart, ItemWeight, ShoppingCartItem, FlatRateShipping, Pickup

__all__ = (
    # These are the objects needed for making requests.
    # The objects that only occur in replies are not included here.
    'Price',
    'Merchant', 'Plugin',
    'Customer', 'CustomerDelivery',
    'GatewayCustomer', 'Gateway',
    'GatewayInfo',
    'Transaction', 'CheckoutSettings', 'GoogleAnalytics',
    'Cart', 'ItemWeight', 'ShoppingCartItem', 'FlatRateShipping', 'Pickup',
)



import hashlib

from .base import XmlRequest, XmlResponse


class RedirectTransaction(XmlRequest):
    """
    The message to start a checkout, using the Connect method.
    """
    xml_name = 'redirecttransaction'
    xml_fields = (
        'merchant',
        'customer',
        'google_analytics',
        'transaction',
        'plugin',
        'signature',
    )

    def __init__(self, merchant, transaction, customer, plugin=None, google_analytics=None):
        """
        :type merchant: Merchant
        :type transaction: Transaction
        :type customer: Customer
        :type google_analytics: GoogleAnalytics
        """
        self.merchant = merchant
        self.transaction = transaction
        self.customer = customer
        self.plugin = plugin
        self.google_analytics = google_analytics

    @property
    def signature(self):
        data = '{0}{1}{2}{3}{4}'.format(
            self.transaction.amount,
            self.transaction.currency,
            self.merchant.account,
            self.merchant.site_id,
            self.transaction.id
        )
        return hashlib.md5(str(data)).hexdigest()


class RedirectTransactionReply(XmlResponse):
    """
    Reply from a redirecttransaction call.
    """

    def __init__(self, id, payment_url):
        """
        :param id: ID of the session
        :param payment_url: The URL to redirect to.
        """
        self.id = id
        self.payment_url = payment_url

    @classmethod
    def get_class_kwargs(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        transaction = xml.find('transaction')
        return dict(
            id=transaction.find('id').text,
            payment_url=transaction.find('payment_url').text,
        )



from .redirecttransaction import RedirectTransaction, RedirectTransactionReply


class CheckoutTransaction(RedirectTransaction):
    """
    The message to start a checkout, using the Fast-Connect method.
    """
    xml_name = 'checkouttransaction'
    xml_fields = (
        'merchant',
        'plugin',
        'customer',
        'customer_delivery',
        'cart',             # TODO: Cart data model not implemented yet   (uses Google Checkout Cart format)
        'fields',           # TODO: Fields data model not implemented yet (for custom form fields)
        'google_analytics',
        'checkout_settings',
        'transaction',
        'signature',
    )

    def __init__(self, merchant, transaction, customer, customer_delivery=None, cart=None, fields=None, plugin=None, checkout_settings=None, google_analytics=None):
        """
        :type merchant: Merchant
        :type transaction: Transaction
        :type customer: Customer
        :type customer_delivery: CustomerDelivery
        :type cart: Cart
        :type plugin: Plugin
        :type google_analytics: GoogleAnalytics
        """
        super(CheckoutTransaction, self).__init__(
            merchant=merchant,
            transaction=transaction,
            customer=customer,
            google_analytics=google_analytics
        )

        # Fast-Checkout fields that are not present in the redirecttransaction of the Connect method.
        self.plugin = plugin
        self.customer_delivery = customer_delivery
        self.cart = cart
        self.fields = fields
        self.checkout_settings = checkout_settings


class CheckoutTransactionReply(RedirectTransactionReply):
    """
    Reply from a start_transaction call.
    Is identical to the Connect method reply.
    """



"""
All exceptions raised by this package.
"""


class MultiSafepayException(Exception):
    """
    Base class for all exceptions from MultiSafepay
    """
    CODE_INVALID_TRANSACTION_ID = '1006'


class MultiSafepayServerException(MultiSafepayException):
    """
    Base class for reported errors via the API.
    """

    def __init__(self, code, description):
        super(MultiSafepayServerException, self).__init__("{0}: {1}".format(code, description))
        self.code = code
        self.description = description

    @classmethod
    def from_xml(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        error = xml.find('error')
        kwargs = dict(
            code=error.find('code').text,
            description=error.find('description').text
        )

        # NOTE: depending on the error code, some custom elements might be available.
        # This is currently not being checked for.
        # We could implement a class type per exception code, if that's needed.

        return cls(**kwargs)



from .base import XmlObject


class Customer(XmlObject):
    """
    Customer information
    """
    xml_name = 'customer'
    xml_fields = (
        'locale',
        'ipaddress',
        'forwardedip',
        'firstname',
        'lastname',
        'address1',
        'address2',
        'housenumber',
        'zipcode',
        'city',
        'state',
        'country',
        'phone',
        'email',
        'referrer',
        'user_agent'
    )

    def __init__(self, locale, firstname, lastname, address1, address2, housenumber, zipcode, city, state, country, phone, email, ipaddress=None, forwardedip=None, referrer=None, user_agent=None):
        """
        :param locale: Customer locale, e.g. AB_cd; the language (ISO 639) and country (ISO 3166-1)
        :param country: The 2-digit country code, in ISO 3166
        """
        self.locale = locale
        self.firstname = firstname
        self.lastname = lastname
        self.address1 = address1
        self.address2 = address2
        self.housenumber = housenumber
        self.zipcode = zipcode
        self.city = city
        self.state = state
        self.country = country
        self.phone = phone
        self.email = email

        self.ipaddress = ipaddress
        self.forwardedip = forwardedip
        self.referrer = referrer
        self.user_agent = user_agent


class CustomerDelivery(XmlObject):
    """
    Delivery address
    """
    xml_name = 'customer-delivery'
    xml_fields = (
        'firstname',
        'lastname',
        'address1',
        'address2',
        'housenumber',
        'zipcode',
        'city',
        'state',
        'country',
        'phone',
        'email',
    )

    @classmethod
    def from_customer(cls, customer):
        """
        Initialize this class using the information of the customer.
        """
        return cls(
            firstname=customer.firstname,
            lastname=customer.lastname,
            address1=customer.address1,
            address2=customer.address2,
            housenumber=customer.housenumber,
            zipcode=customer.zipcode,
            city=customer.city,
            state=customer.state,
            country=customer.country,
            phone=customer.phone,
            email=customer.email,
        )

    def __init__(self, firstname, lastname, address1, address2, housenumber, zipcode, city, state, country, phone, email):
        """
        :param country: The 2-digit country code, in ISO 3166
        """
        self.firstname = firstname
        self.lastname = lastname
        self.address1 = address1
        self.address2 = address2
        self.housenumber = housenumber
        self.zipcode = zipcode
        self.city = city
        self.state = state
        self.country = country
        self.phone = phone
        self.email = email



from .base import XmlObject


class GatewayInfo(XmlObject):
    """
    <gatewayinfo>
        <issuerid>0151</issuerid>
    </gatewayinfo>
    """
    xml_name = 'gatewayinfo'
    xml_fields = (
        'issuerid',
    )

    def __init__(self, issuerid):
        self.issuerid = issuerid



from django.core.urlresolvers import reverse
from django_multisafepay import PLUGIN_VERSION, SHOP_NAME, appsettings

from .base import XmlObject


class Merchant(XmlObject):
    """
    Meta information for the webshop.
    """
    xml_name = 'merchant'
    xml_fields = (
        'account',
        'site_id',
        'site_secure_code',
        'notification_url',
        'cancel_url',
        'redirect_url',
        'close_window',
    )

    def __init__(self, account=None, site_id=None, site_code=None, notification_url=None, cancel_url=None, redirect_url=None, close_window=None):
        """
        If the account, site_id and site_code are omitted, the defaults are read from the settings file.
        """
        self.account = account or appsettings.MULTISAFEPAY_ACCOUNT_ID
        self.site_id = site_id or appsettings.MULTISAFEPAY_SITE_ID
        self.site_secure_code = site_code or appsettings.MULTISAFEPAY_SITE_CODE

        # Custom configuration
        self.notification_url = notification_url or reverse('notification_url')
        self.cancel_url = cancel_url or appsettings.MULTISAFEPAY_CANCEL_URL
        self.redirect_url = redirect_url or appsettings.MULTISAFEPAY_REDIRECT_URL
        self.close_window = close_window


class Plugin(XmlObject):
    """
    Meta information for the plugin
    """
    xml_name = 'plugin'
    xml_fields = (
        'shop',
        'shop_version',
        'plugin_version',
        'partner',
        'shop_root_url',
    )

    def __init__(self, shop=None, shop_version=None, plugin_version=None, partner=None, shop_root_url=None):
        """
        :param shop: Name of the webshop software (defaults to "django-multisafepay")
        :param shop_version: Version of the webshop software.
        :param plugin_version: Version of this plugin (defaults to current package version)
        """
        self.shop = shop or SHOP_NAME
        self.shop_version = shop_version
        self.plugin_version = plugin_version or PLUGIN_VERSION
        self.partner = partner
        self.shop_root_url = shop_root_url



from .base import XmlObject, escape


class Transaction(XmlObject):
    """
    Transaction data
    """
    xml_name = 'transaction'
    xml_fields = (
        'id',
        'currency',
        'amount',
        'description',
        'var1',    # custom var 1
        'var2',    # custom var 2
        'var3',    # custom var 3
        'items',
        'manual',
        'daysactive',
        # 'gateway' added manually
    )

    def __init__(self, id, currency=None, amount=None, description=None, items=None, manual=None, daysactive=None, gateway=None, gateway_issuer=None, var1=None, var2=None, var3=None):
        """
        :param currency: Currency, e.g. "EUR", "USD"
        :param amount: Value in cents
        :type amount: int
        :param items: HTML text for items, e.g. ``'<br/><ul><li>1 x Item1</li><li>2 x Item2</li></ul>'``.
        :param items: str
        :param manual: Typically ``False``.
        :type manual: bool
        :param var1: Custom variable, to be returned at the callback.
        :param var2: Custom variable, to be returned at the callback.
        :param var3: Custom variable, to be returned at the callback.
        """
        self.id = id
        self.currency = currency
        self.amount = int(amount) if amount else None
        self.description = description
        self.items = items
        self.manual = manual
        self.daysactive = daysactive
        self.var1 = var1
        self.var2 = var2
        self.var3 = var3
        self.gateway = gateway
        self.gateway_issuer = gateway_issuer

    def get_xml_children(self):
        lines = super(Transaction, self).get_xml_children()
        if self.gateway is not None:
            # Add issuer attribute to the gateway tag
            issuer = u' issuer="{0}"'.format(escape(self.gateway_issuer)) if self.gateway_issuer else u''
            lines.append(u'<gateway{0}>{1}</gateway>'.format(issuer, escape(self.gateway)))
        return lines


class CheckoutSettings(XmlObject):
    """"
    Passing checkout sessions to the :class:`CheckoutTransaction` class.
    """
    xml_name = 'checkout-settings'
    xml_fields = (
        'use-shipping-notification',
    )

    def __init__(self, use_shipping_notification):
        super(CheckoutSettings, self).__init__()
        self.use_shipping_notification = bool(use_shipping_notification)


class GoogleAnalytics(XmlObject):
    """
    Passing analytics tracker to the payment pages.
    """
    xml_name = 'google-analytics'
    xml_fields = (
        'account',
    )

    def __init__(self, account):
        super(GoogleAnalytics, self).__init__()
        self.account = account



from django.utils.translation import to_locale

from .base import XmlObject


class GatewayCustomer(XmlObject):
    """
    Customer information, for the Gateways request
    """
    xml_name = 'customer'
    xml_fields = (
        'locale',
        'country',
    )

    def __init__(self, locale, country):
        """
        :param locale: Language code, e.g. en_US
        :param country: The 2-digit country code, in ISO 3166
        """
        self.locale = to_locale(locale)  # ensure xx_XX format
        self.country = country


class Gateway(XmlObject):
    """
    <gateway>
        <id>IDEAL</id>
        <description>iDEAL</description>
    </gateway>
    """
    xml_name = 'gateway'
    xml_fields = (
        'id',
        'description',
    )



from django_multisafepay.data.gateway import Gateway

from .base import XmlRequest, XmlResponse


class Gateways(XmlRequest):
    """
    The message to start a checkout
    """
    xml_name = 'gateways'
    xml_fields = (
        'merchant',
        'customer',
    )

    def __init__(self, merchant, customer):
        """
        :type merchant: Merchant
        :type customer: GatewayCustomer
        """
        self.merchant = merchant
        self.customer = customer


class GatewaysReply(XmlResponse):
    """
    Reply from a start_transaction call.
    """

    def __init__(self, gateways):
        """
        :param id: ID of the session
        :param payment_url: The URL to redirect to.
        """
        self.gateways = list(gateways)

    def __iter__(self):
        return iter(self.gateways)

    @classmethod
    def get_class_kwargs(cls, xml):
        """
        Split the object to all ``__init__()`` parameters.
        :type xml: xml.etree.ElementTree.Element
        """
        gateways = xml.find('gateways')
        return dict(
            gateways=[Gateway.from_xml(gateway) for gateway in gateways]
        )



from xml.etree import ElementTree

from django_multisafepay import USER_AGENT
from django_multisafepay.data.base import XmlObject


class XmlRequest(XmlObject):
    """
    A root XML node.
    """

    def to_xml(self):
        lines = self.get_xml_children()
        return u'<?xml version="1.0" encoding="UTF-8"?>\n' \
               u'<{0} ua="{1}">{2}</{0}>'.format(self.xml_name, USER_AGENT, u''.join(lines))


class XmlResponse(object):
    """
    Base class for response objects.
    """
    _xml = None

    @classmethod
    def from_xml(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        if xml is None:
            return None
        kwargs = cls.get_class_kwargs(xml)  # Make kwargs available in debugging stack frame.
        reply = cls(**kwargs)
        reply._xml = xml  # Inject response for better logging in Sentry.
        return reply

    @classmethod
    def get_class_kwargs(cls, xml):
        """
        Custom method to override, provide all ``__init__()`` fields by parsing the XML.
        :param xml: The incoming XML message
        :type xml: xml.etree.ElementTree.Element
        :return: The parameters for the init method.
        :rtype: dict
        """
        return {}

    def __repr__(self):
        if self._xml is not None:
            return '<{0} {1}>'.format(self.__class__.__name__, ElementTree.tostring(self._xml, encoding='utf8'))
        else:
            return '<{0} without _xml data>'.format(self.__class__.__name__)



from .redirecttransaction import RedirectTransaction, RedirectTransactionReply


class DirectTransaction(RedirectTransaction):
    """
    The message to start a checkout and skip the initial choices, using the Fast-Connect method.
    """
    xml_name = 'directtransaction'
    xml_fields = RedirectTransaction.xml_fields + (
        'gatewayinfo',
    )

    def __init__(self, merchant, transaction, customer, gatewayinfo, google_analytics=None):
        """
        :type merchant: Merchant
        :type transaction: Transaction
        :type customer: Customer
        :type gatewayinfo: GatewayInfo
        :type google_analytics: GoogleAnalytics
        """
        super(DirectTransaction, self).__init__(
            merchant=merchant,
            transaction=transaction,
            customer=customer,
            google_analytics=google_analytics
        )

        self.gatewayinfo = gatewayinfo


class DirectTransactionReply(RedirectTransactionReply):
    """
    Reply from a directtransaction call.
    Is identical to the standard method reply.
    """



"""
Settings for this application
"""
from django.conf import settings

MULTISAFEPAY_ACCOUNT_ID = getattr(settings, 'MULTISAFEPAY_ACCOUNT_ID')
MULTISAFEPAY_SITE_ID = getattr(settings, 'MULTISAFEPAY_SITE_ID')
MULTISAFEPAY_SITE_CODE = getattr(settings, 'MULTISAFEPAY_SITE_CODE')

# Whether to use the testing mode, or live mode.
MULTISAFEPAY_TESTING = getattr(settings, 'MULTISAFEPAY_TESTING', True)

# Allow to define the URLs globally (e.g. using reverse_lazy())
MULTISAFEPAY_REDIRECT_URL = getattr(settings, 'MULTISAFEPAY_REDIRECT_URL', None)
MULTISAFEPAY_CANCEL_URL = getattr(settings, 'MULTISAFEPAY_CANCEL_URL', None)



from django_multisafepay.data import Customer

from .base import Price, XmlObject


class Ewallet(XmlObject):
    """
    The EWallet element in the status reply.
    It contains the status code.
    """
    xml_name = 'ewallet'
    xml_fields = (
        'id',
        'fastcheckout',
        'status',
        'created',
        'modified',
        'reason',
        'reasoncode',
    )

    def __init__(self, id, status, fastcheckout, created, modified, reason=None, reasoncode=None):
        self.id = id
        self.status = status
        self.fastcheckout = bool(fastcheckout is True or fastcheckout == 'YES')
        self.created = created
        self.modified = modified
        self.reason = reason
        self.reasoncode = reasoncode


class CustomerStatus(Customer):
    """
    A customer with additional fields.
    """
    xml_fields = Customer.xml_fields + (
        'amount',
        'currency',
        'account',
        'phone1',
        'phone2',
        'countryname'
    )

    def __init__(self, *args, **kwargs):
        self.amount = kwargs.pop('amount')
        self.currency = kwargs.pop('currency')
        self.account = kwargs.pop('account')
        self.phone1 = kwargs.pop('phone1')
        self.phone2 = kwargs.pop('phone2')
        self.countryname = kwargs.pop('countryname')
        super(CustomerStatus, self).__init__(*args, **kwargs)


class PaymentDetails(XmlObject):
    """
    The payment details in the status reply
    """
    xml_name = 'paymentdetails'
    xml_fields = (
        'type',
        'accountid',
        'accountholdername',
        'externaltransactionid',
    )

    def __init__(self, type, accountid, accountholdername, externaltransactionid):
        self.type = type
        self.accountid = accountid
        self.accountholdername = accountholdername
        self.externaltransactionid = externaltransactionid


class CheckoutData(XmlObject):
    """
    The checkout data in the status reply
    """
    xml_name = 'checkoutdata'
    xml_fields = (
        'shopping-cart',
        'order-adjustment',
        'order-total',
        'custom-fields',
    )

    def __init__(self, order_total, shopping_cart=None, order_adjustment=None, custom_fields=None):
        self.order_total = order_total
        self.shopping_cart = shopping_cart
        self.order_adjustment = order_adjustment
        self.custom_fields = custom_fields

    @classmethod
    def get_class_kwargs(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        return dict(
            order_total=Price.from_xml(xml.find('order-total')),
            order_adjustment=OrderAdjustment.from_xml(xml.find('order-adjustment')),
            shopping_cart=None,
            custom_fields=None,
            # TODO: all other objects are currently ignored. (shopping-cart, custom-fields)
        )


class TransactionStatus(XmlObject):
    """
    Information about a transaction, as returned by the status call..
    """
    xml_name = 'transaction'
    xml_fields = (
        'id',
        'recurringid',
        'currency',
        'amount',
        'cost',
        'description',
        'var1',    # custom var 1
        'var2',    # custom var 2
        'var3',    # custom var 3
        'items',
        'amountrefunded',
    )

    def __init__(self, id, recurringid, currency, amount, cost, description, var1=None, var2=None, var3=None, items=None, amountrefunded=None):
        self.id = id
        self.recurringid = recurringid
        self.currency = currency
        self.amount = amount
        self.cost = cost
        self.description = description
        self.var1 = var1
        self.var2 = var2
        self.var3 = var3
        self.items = items
        self.amountrefunded = amountrefunded


class OrderAdjustment(XmlObject):
    """
    The order adjustment in the CheckoutData of the status reply.
    """
    xml_name = 'order-adjustment'
    xml_fields = (
        'shipping',
        'adjustment-total',
        'total-tax'
    )

    def __init__(self, shipping, adjustment_total, total_tax):
        self.shipping = shipping
        self.adjustment_total = adjustment_total
        self.total_tax = total_tax

    @classmethod
    def get_class_kwargs(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        return dict(
            shipping=None,   # TODO: not implemented
            adjustment_total=Price.from_xml(xml.find('adjustment-total')),
            total_tax=Price.from_xml(xml.find('total-tax')),
        )



from decimal import Decimal
from xml.sax.saxutils import escape

from django.utils.encoding import force_text

__all__ = (
    'XmlObject',
    'Price',
    'escape',
)


class XmlObject(object):
    """
    Simple object to quickly generate XML messages.
    This provides all the flexibility we need to render any XML object.
    """
    xml_name = None
    xml_attrs = None
    xml_fields = ()

    def __repr__(self):
        return "<{0} {1}>".format(self.__class__.__name__, self.to_xml().encode('utf-8'))

    def to_xml(self):
        # get xml message
        lines = self.get_xml_children()
        attrs = self.get_xml_attrs()
        if attrs:
            attrs = u"".join(u' {0}="{1}"'.format(k, escape(force_text(v))) for k, v in attrs.iteritems())
            return u'<{0}{1}>{2}</{0}>'.format(self.xml_name, attrs, u''.join(lines))
        else:
            return u'<{0}>{1}</{0}>'.format(self.xml_name, u''.join(lines))

    def get_xml_attrs(self):
        return self.xml_attrs

    def get_xml_children(self):
        # Allow to be overwritten
        lines = []
        for field in self.xml_fields:
            value = getattr(self, field.replace('-', '_'))
            if value is not None:
                if isinstance(value, XmlObject):
                    # Attribute name is ignored, tag name is used instead.
                    lines.append(u"{0}\n".format(value.to_xml()))
                elif isinstance(value, Price):
                    # Inconsistent API. Using decimal notation here, but using cents somewhere else.
                    lines.append(u'<{0} currency="{1}">{2:.2f}</{0}>'.format(field, value.currency, value))
                else:
                    if isinstance(value, (list, tuple)):
                        tag_value = u'\n'.join(item.to_xml() for item in value)
                    elif isinstance(value, bool):
                        tag_value = str(value).lower()
                    else:
                        tag_value = escape(force_text(value))
                    lines.append(u'<{0}>{1}</{0}>'.format(field, tag_value))
        return lines

    @classmethod
    def from_xml(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        if xml is None:
            return None
        kwargs = cls.get_class_kwargs(xml)  # Make kwargs available in debugging stack frame.
        return cls(**kwargs)

    @classmethod
    def get_class_kwargs(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        kwargs = {}
        for field in cls.xml_fields:
            node = xml.find(field)
            kwargs[field.replace('-', '_')] = None if node is None else node.text
        return kwargs


class Price(Decimal):
    """
    A decimal value with currency attached.
    """

    def __new__(cls, value, currency=None):
        self = Decimal.__new__(cls, value)
        self.currency = currency
        return self

    @classmethod
    def from_xml(cls, xml):
        """
        :type xml: xml.etree.ElementTree.Element
        """
        return cls(xml.text, xml.attrib['currency'])



from .base import XmlObject, escape


class Cart(XmlObject):
    """
    The shopping cart.
    """
    xml_name = 'checkout-shopping-cart'

    def __init__(self, items=(), shipping_methods=(), tax_tables=()):
        """
        :type items: list of :class:`Item`
        """
        self.items = list(items or ())
        self.shipping_methods = list(shipping_methods or ())
        self.tax_tables = list(tax_tables or ())

    def get_xml_children(self):
        lines = []
        if self.items:
            # Don't bother creating sub elements for the shopping cart, make this object smarter.
            lines.append('<shopping-cart><items>\n')
            for item in self.items:
                lines.append("  {0}\n".format(item.to_xml()))
            lines.append('</items></shopping-cart>\n')

        if self.shipping_methods or self.tax_tables:
            lines.append('<checkout-flow-support><merchant-checkout-flow-support>\n')

            if self.shipping_methods:
                lines.append('<shipping-methods>\n')
                for item in self.shipping_methods:
                    lines.append("  {0}\n".format(item.to_xml()))
                lines.append('</shipping-methods>\n')

            if self.tax_tables:
                lines.append('<tax-tables>\n')
                for item in self.tax_tables:
                    lines.append("  {0}\n".format(item.to_xml()))
                lines.append('</tax-tables>\n')

            lines.append('</merchant-checkout-flow-support></checkout-flow-support>\n')
        return lines


class ShoppingCartItem(XmlObject):
    """
    An item in the shopping cart.
    """
    xml_name = 'item'
    xml_fields = (
        'item-name',
        'item-description',
        'unit-price',
        'quantity',
        'merchant-item-id'
        'item-weight'  # unit="KG" value="1"
    )

    def __init__(self, item_name, item_description, unit_price, quantity, merchant_item_id, item_weight=None):
        """
        :type unit_price: Price
        :type item_weight: ItemWeight
        """
        self.item_name = item_name
        self.item_description = item_description
        self.unit_price = unit_price
        self.quantity = quantity
        self.merchant_item_id = merchant_item_id
        self.item_weight = item_weight


class ItemWeight(XmlObject):
    """
    Weight value for an item in the shopping cart.
    """
    xml_name = 'item-weight'

    def __init__(self, value, unit):
        """
        :param value: The actual value.
        :param unit: The units, e.g. "KG".
        """
        self.value = value
        self.unit = unit

    def to_xml(self):
        return u'<{0} unit="{1}" value="{2}" />'.format(self.xml_name, escape(self.unit), escape(str(self.value)))


class ShippingMethodBase(XmlObject):
    """
    Base class for shipping methods
    """
    xml_fields = (
        'price',
    )

    def __init__(self, name, price):
        """
        :type name: str
        :type price: Price
        """
        self.name = name
        self.price = price

    def get_xml_attrs(self):
        return {
            'name': self.name
        }


class Pickup(ShippingMethodBase):
    """
    Shipping costs for pickup.
    """
    xml_name = 'pickup'


class FlatRateShipping(ShippingMethodBase):
    """
    Shipping costs for flat-rates
    """
    xml_name = 'flat-rate-shipping'
    xml_fields = ShippingMethodBase.xml_fields + (
        'shipping_restrictions',  # TODO: data class is not implemented.
    )

    def __init__(self, name, price, restrictions=None):
        """
        :type name: str
        :type price: Price
        :type restrictions
        :param restrictions: Not implemented yet.
        """
        super(FlatRateShipping, self).__init__(name, price)
        self.restrictions = restrictions



import django

# following PEP 440
__version__ = "1.1"

SHOP_NAME = "django-multisafepay"
PLUGIN_VERSION = __version__

USER_AGENT = '{0}/{1} (Django {2}.{3}.x)'.format(SHOP_NAME, PLUGIN_VERSION, django.VERSION[0], django.VERSION[1])



"""
API calls to the payment gateway webservice
"""
import logging
from xml.etree import ElementTree

import requests
from django_multisafepay import __version__ as package_version
from django_multisafepay import appsettings, messages
from django_multisafepay.data import Merchant, Plugin
from django_multisafepay.data.gateway import GatewayCustomer
from django_multisafepay.exceptions import MultiSafepayException, MultiSafepayServerException

logger = logging.getLogger(__name__)


URL_TEST = "https://testapi.multisafepay.com/ewx/"
URL_LIVE = "https://api.multisafepay.com/ewx/"


class MultiSafepayClient(object):
    """
    The MultiSafepay API client.
    """

    def __init__(self, merchant=None, plugin=None, is_test=None):
        """
        Provide account details to call the service.

        :param merchant: Custom merchant settings. By default the settings file is read.
        :type merchant: Merchant
        :param plugin: Custom metadata about the shop-plugin that is used.
        :type plugin: Plugin
        :param is_test: Whether to use testing mode. Using ``None`` defaults to the defined setting value.
        :type is_test: bool
        """
        self.merchant = merchant or Merchant()
        self.plugin = plugin or Plugin()
        self.is_test = is_test if is_test is not None else appsettings.MULTISAFEPAY_TESTING

    @property
    def api_url(self):
        return URL_TEST if self.is_test else URL_LIVE

    def _call(self, message, response_class=None):
        """
        Make the call to the server
        :rtype: :class:`xml.etree.ElementTree.Element` | response_class
        """
        postdata = message.to_xml()
        logger.debug(u"sending to {0}:\n{1}".format(self.api_url, postdata))

        response = requests.post(
            self.api_url,
            headers={
                'Content-Type': 'text/xml',
                'Accept': 'text/xml; charset=utf-8',
                'User-Agent': 'django-multisafepay/{0}'.format(package_version),
            },
            data=postdata.encode('utf-8'),
            allow_redirects=False,
            verify=True
        )

        # Raise on invalid status
        if response.status_code >= 400:
            logger.error(u"http failed: {0} {1}".format(response.status_code, response.content))
            response.raise_for_status()

        if '/xml' not in response.headers['Content-Type']:
            raise MultiSafepayException("Received invalid content-type: {0}".format(response.headers['Content-Type']))

        # Fix MultiSafepay response header error.
        # Encoding is only specified in the <?xml preamble, not in the HTTP Content-Type header.
        # "response.text" parses the response in the proper encoding format.
        # "response.content" is the raw content, that's being sent to the XML parser.
        response.encoding = 'utf-8'
        logger.debug(u"http succeeded: {0}".format(response.text))

        xml = ElementTree.fromstring(response.content)  # parser=ElementTree.XMLParser(encoding=response.encoding))  # Python 2.6 doesn't support this.

        if xml.attrib['result'] != 'ok':
            ex = MultiSafepayServerException.from_xml(xml)
            logger.error(u"Request <%s> to MultiSafePay failed: code=%s, description=%s", message.xml_name, ex.code, ex.description)
            raise ex

        if response_class is not None:
            return response_class.from_xml(xml)
        else:
            return xml

    def _transaction_call(self, message, response_class=None):
        """
        A variant of the standard ``_call()`` that logs the transaction ID on errors.

        :type message: RedirectTransaction
        :type response_class: RedirectTransactionReply
        :rtype: RedirectTransactionReply
        """
        try:
            return self._call(message, response_class=response_class)
        except MultiSafepayServerException as e:
            # Be more verbose in the logs.
            logger.error(u"Failed to start transaction %s: code=%s, description=%s", message.transaction.id, e.code, e.description)
            if e.code == e.CODE_INVALID_TRANSACTION_ID:
                # Mention transaction ID in exception message
                raise MultiSafepayServerException(e.code, u"{0} ({1})".format(e.description, message.transaction.id))
            raise

    def start_checkout(self, transaction, customer, customer_delivery=None, cart=None, fields=None, checkout_settings=None, google_analytics=None):
        """
        Start the checkout (Fast-Checkout method)

        :param transaction: The information about the transaction.
        :type transaction: Transaction
        :param customer: The information about the customer.
        :type customer: Customer
        :param customer_delivery: The delivery address.
        :type customer_delivery: CustomerDelivery
        :param cart: The shopping cart.
        :type cart: Cart
        :param fields: Not supported yet.
        :param checkout_settings: Checkout sessions, only provides "use shipping notification" for now.
        :type checkout_settings: CheckoutSettings
        :param google_analytics: Analytics to use on the payment page.
        :type google_analytics: GoogleAnalytics
        :rtype: CheckoutTransactionReply
        """
        request = messages.CheckoutTransaction(
            merchant=self.merchant,
            transaction=transaction,
            customer=customer,
            customer_delivery=customer_delivery,
            cart=cart,
            fields=fields,
            plugin=self.plugin,
            checkout_settings=checkout_settings,
            google_analytics=google_analytics
        )

        return self._transaction_call(request, response_class=messages.CheckoutTransactionReply)

    def status(self, transaction_id):
        """
        Request the status of a transaction.
        :rtype: StatusReply
        """
        request = messages.Status(self.merchant, transaction_id)
        try:
            return self._call(request, response_class=messages.StatusReply)
        except MultiSafepayServerException as e:
            # Be more verbose in the logs.
            logger.error(u"Failed to fetch status for transaction %s: code=%s, description=%s", transaction_id, e.code, e.description)
            raise

    def gateways(self, locale, country):
        """
        Request all available gateways (Connect method)

        :param locale: Language code, e.g. en_US
        :param country: The 2-digit country code, in ISO 3166
        :rtype: GatewaysReply
        """
        request = messages.Gateways(
            merchant=self.merchant,
            customer=GatewayCustomer(
                locale=locale,
                country=country,
            )
        )

        return self._call(request, response_class=messages.GatewaysReply)

    def redirect_transaction(self, transaction, customer, google_analytics=None):
        """
        Start the checkout (Connect method)

        :param transaction: The information about the transaction.
        :type transaction: Transaction
        :param customer: The information about the customer.
        :type customer: Customer
        :param google_analytics: Analytics to use on the payment page.
        :type google_analytics: GoogleAnalytics
        :rtype: CheckoutTransactionReply
        """
        request = messages.RedirectTransaction(
            merchant=self.merchant,
            transaction=transaction,
            customer=customer,
            plugin=self.plugin,
            google_analytics=google_analytics
        )

        return self._transaction_call(request, response_class=messages.RedirectTransactionReply)

    def direct_transaction(self, transaction, customer, gatewayinfo, google_analytics=None):
        """
        Start the checkout (Connect method)

        :param transaction: The information about the transaction.
        :type transaction: Transaction
        :param customer: The information about the customer.
        :type customer: Customer
        :param gatewayinfo: The exact payment gateway to use.
        :type gatewayinfo: GatewayInfo
        :param google_analytics: Analytics to use on the payment page.
        :type google_analytics: GoogleAnalytics
        :rtype: CheckoutTransactionReply
        """
        request = messages.DirectTransaction(
            merchant=self.merchant,
            transaction=transaction,
            customer=customer,
            gatewayinfo=gatewayinfo,
            google_analytics=google_analytics
        )

        return self._transaction_call(request, response_class=messages.DirectTransactionReply)

    # TODO: idealissuers request



import logging

from django.db import transaction
from django.http import HttpResponse
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from django.views.generic import View
from django_multisafepay.client import MultiSafepayClient
from django_multisafepay.signals import order_status_updated

logger = logging.getLogger(__name__)

# New transaction support in Django 1.6
try:
    transaction_atomic = transaction.atomic
except AttributeError:
    transaction_atomic = transaction.commit_on_success


class NotificationView(View):
    """
    View to be called by MultiSafepay when a status update occured.
    """
    return_link_text = _('Return to webshop')
    client_class = MultiSafepayClient

    def get(self, request, *args, **kwargs):
        """
        Handle the incoming notification request.
        """
        try:
            self.transaction_id = request.GET['transactionid']
        except KeyError:
            return HttpResponse("missing transactionid", status=403)
        self.type = request.GET.get('type')

        # Request the new status from the server.
        client = self.get_client()
        statusreply = client.status(self.transaction_id)

        # Let the project update the status
        if not order_status_updated.has_listeners():
            logger.warning("No listeners for `order_status_updated` signal!")
        else:
            with transaction_atomic():
                order_status_updated.send(self.__class__, statusreply=statusreply, request=self.request)

        if self.type == 'initial':
            # displayed at the last page of the transaction process (if no redirect_url is set)
            return self.render_to_response()
        else:
            # MultiSafepay back-end expects an "ok" if no error occurred
            return HttpResponse('ok')

    def get_client(self):
        """
        Return the MultiSafepay API client.
        This method can be overwritten to create a custom client, with other merchant parameters for example.
        """
        return self.client_class()

    def render_to_response(self):
        """
        Render the response when no redirect_url is set.
        """
        url = self.request.build_absolute_uri('/')
        return HttpResponse(format_html(u'<a href="{0}">{1}</a>', url, self.return_link_text))



#!/usr/bin/env python
from setuptools import setup, find_packages
from os import path
import codecs
import os
import re
import sys


def read(*parts):
    file_path = path.join(path.dirname(__file__), *parts)
    return codecs.open(file_path, encoding='utf-8').read()


def find_version(*parts):
    version_file = read(*parts)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return str(version_match.group(1))
    raise RuntimeError("Unable to find version string.")


setup(
    name='django-multisafepay',
    version=find_version('django_multisafepay', '__init__.py'),
    license='Apache 2.0',

    install_requires = [
        'requests>=1.2.1'
    ],
    requires=[
        'Django (>=1.4)',
    ],

    description='MultiSafepay Payments Gateway integration for Django',
    long_description=read('README.rst'),

    author='Diederik van der Boor',
    author_email='opensource@edoburu.nl',

    url='https://github.com/edoburu/django-multisafepay',
    download_url='https://github.com/edoburu/django-multisafepay/zipball/master',

    packages=find_packages(exclude=('example*',)),
    include_package_data=True,

    #test_suite = 'runtests',

    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Office/Business :: Financial',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



"""
The messages are mainly used for internal purposes.
The describe the XML format for the calls that the MultiSafepay client makes.
"""
from .checkouttransaction import CheckoutTransaction, CheckoutTransactionReply
from .directtransaction import DirectTransaction, DirectTransactionReply
from .status import Status, StatusReply
from .gateway import Gateways, GatewaysReply
from .redirecttransaction import RedirectTransaction, RedirectTransactionReply

__all__ = (
    'CheckoutTransaction',
    'CheckoutTransactionReply',

    'Status',
    'StatusReply',

    'Gateways',
    'GatewaysReply',

    'RedirectTransaction',
    'RedirectTransactionReply',

    'DirectTransaction',
    'DirectTransactionReply',
)
